// C/C++ Headers

#include <stdio.h>

// Qt Headers

#include <qstring.h>

// Local Headers

#include "qscalereader.h"
#include "qtagvisitor.h"
#include "qtag.h"

/*****************************************************************************
 * local class: ValidationVisitor
 *****************************************************************************/

class ValidationVisitor : public QTagVisitor
{
   public:
      ValidationVisitor(int patternLength);
      ~ValidationVisitor();
      void begin(const QTag *theTag);
      QString &getErrMsg();

   private:
      ValidationVisitor(const ValidationVisitor &);
      ValidationVisitor &operator=(const ValidationVisitor &);

      int mPatternLength;
      QString mErrMsg;
      int mCount;
};

ValidationVisitor::ValidationVisitor(int patternLength) :
   mPatternLength(patternLength),
   mErrMsg(""),
   mCount(0)
{
}

ValidationVisitor::~ValidationVisitor()
{
}

void ValidationVisitor::begin(const QTag *theTag)
{
   // If this is the first tag, we expect it to be the top-level
   // tag QSCALE

   QString msg("");
   const QString &tagName = theTag->name();

   if(mCount == 0)
   {
      if(tagName != QSCALE)
      {
         msg.sprintf(
            "Expected top level tag to be: [%s]",
            QSCALE
         );
      }
   }
   else
   {
      // Ok, not the top-level tag - the name should be ENTRY

      if(tagName != ENTRY)
      {
         msg.sprintf(
            "Invalid tag name: [%s]",
            tagName.latin1()
         );
      }
      else
      {
         // Alright, what sort of ENTRY tag is this?

         if(!theTag->getAttribute(NAME).isNull())
         {
            // This is a "folder" tag, ensure it has at least one
            // child

            if(theTag->getChildCount() == 0)
            {
               msg.sprintf(
                  "Folder tag: [%s] has no children",
                  theTag->getAttribute(NAME).latin1()
               );
            }
         }
         else
         {
            // This should be a scale entry

            if(theTag->getAttribute(SCALE).isNull() ||
               theTag->getAttribute(PATTERN).isNull())
            {
               msg.sprintf(
                  "%s tags must either have the: [%s] attribute or "
                  "both the: [%s] and: [%s] attributes",
                  ENTRY, NAME, SCALE, PATTERN
               );
            }
            else
            {
               // Last check, the PATTERN attribute must be patternLength
               // characters long

               const QString &pattern = theTag->getAttribute(PATTERN);

               if((int)pattern.length() != mPatternLength)
               {
                  msg.sprintf(
                     "%s attribute values should be: [%d] chars long (%s)",
                     PATTERN, mPatternLength, pattern.latin1()
                  );
               }
            }
         }
      }
   }

   if(msg.length() > 0)
   {
      if(mErrMsg != "")
         mErrMsg.append("\n");

      mErrMsg.append(msg);
   }

   mCount++;
}

QString &ValidationVisitor::getErrMsg()
{
   return(mErrMsg);
}

/*****************************************************************************
 * definition of class: QScaleReader
 *****************************************************************************/

QTag *QScaleReader::read(
   const char *fileName,
   int patternLength,
   QString &errMsg
)
{
   QString xmlFile(fileName);

   QTag *theTag = QTag::read(xmlFile, errMsg);

   if(theTag)
   {
      // Visit all entries inside 'theTag' to verify that they are valid

      ValidationVisitor visitor(patternLength);

      theTag->accept(visitor);

      QString &error = visitor.getErrMsg();

      if(error != "")
      {
         errMsg = error;
         delete theTag;
         theTag = NULL;
      }
   }

   return(theTag);
}

QTag *QScaleReader::getDefault()
{
   QTag *theTag = new QTag(QSCALE);

   QTag *child = new QTag(ENTRY);

   child->addAttribute(SCALE,   "chromatic");
   child->addAttribute(PATTERN, "XXXXXXXXXXXX");

   theTag->addChild(child);

   return(theTag);
}
