// C/C++ Headers

#include <stdio.h>

// Qt Headers

#include <qtextstream.h>
#include <qstring.h>
#include <qstack.h>
#include <qmap.h>
#include <qfile.h>
#include <qxml.h>


// Local Headers

#include "qtag.h"
#include "qtagvisitor.h"

/*****************************************************************************
 * Local Typedefs
 *****************************************************************************/

typedef QStack<QTag>           QTagStack;
typedef QMap<QString, QString> StringMap;

/*****************************************************************************
 * Local Struct: QTag::Internal
 *****************************************************************************/

struct QTag::Internal
{
   Internal(const QString &tagName);
   ~Internal();

   void dump(FILE *out, const QString &indent = "");

   const QString &getAttribute(const QString &key) const;
   bool hasAttribute(const QString &key) const;
   void getKeys(AttributeKeyList &theList) const;

   bool addAttribute(const QString &key, const QString &value);
   bool removeAttribute(const QString &key);

   int getChildCount() const;
   QTag *getChild(int index) const;
   void addChild(QTag *theChild);
   bool removeChild(int index);
   QTag *takeChild(int index);

   QString   mTagName;
   StringMap mAttributes;
   QTag      **mChildren;
   int       mChildCount;
};

QTag::Internal::Internal(const QString &tagName) :
   mTagName(tagName.lower()),
   mChildren(0),
   mChildCount(0)
{
}

QTag::Internal::~Internal()
{
   for(int i = 0; i < mChildCount; i ++)
   {
      delete mChildren[i];
      mChildren[i] = NULL;
   }

   delete [] mChildren;
   mChildren = NULL;
}

void QTag::Internal::dump(FILE *out, const QString &indent)
{
   QString nextIndent(indent);
   nextIndent.append(" ");

   const char *tagName = mTagName.latin1();
   const char *spaces  = indent.latin1();

   AttributeKeyList theList;
   getKeys(theList);

   int nChildren      = getChildCount();
   const char *suffix = (nChildren == 0 ? "/" : "");

   if(theList.count() == 0)
      fprintf(out, "%s<%s%s>\n", spaces, tagName, suffix);

   else
   {
      fprintf(out, "%s<%s", spaces, tagName);

      AttributeKeyList::Iterator start = theList.begin();
      AttributeKeyList::Iterator end   = theList.end();

      while(start != end)
      {
         const char *key   = (*start).latin1();
         const char *value = getAttribute(*start).latin1();

         fprintf(out, " %s=\"%s\"", key, value);
         start++;
      }

      fprintf(out, "%s>\n", suffix);
   }

   if(nChildren > 0)
   {
      for(int i = 0; i < nChildren; i ++)
         getChild(i)->mInternal->dump(out, nextIndent);

      fprintf(out, "%s</%s>\n", spaces, tagName);
   }
}

const QString &QTag::Internal::getAttribute(const QString &key) const
{
   if(!hasAttribute(key))
      return(QString::null);

   return(mAttributes[key.lower()]);
}

bool QTag::Internal::hasAttribute(const QString &key) const
{
   return(mAttributes.contains(key.lower()));
}

void QTag::Internal::getKeys(AttributeKeyList &theList) const
{
   StringMap::ConstIterator start = mAttributes.begin();
   StringMap::ConstIterator end   = mAttributes.end();

   while(start != end)
   {
      theList.append(start.key());
      start++;
   }
}

bool QTag::Internal::addAttribute(const QString &key, const QString &value)
{
   if(hasAttribute(key))
      return(false);

   mAttributes.insert(key.lower(), value);

   return(true);
}

bool QTag::Internal::removeAttribute(const QString &key)
{
   if(!hasAttribute(key))
      return(false);

   mAttributes.remove(key.lower());

   return(true);
}

int QTag::Internal::getChildCount() const
{
   return(mChildCount);
}

QTag *QTag::Internal::getChild(int index) const
{
   return(index < 0 || index >= mChildCount ? 0 : mChildren[index]);
}

void QTag::Internal::addChild(QTag *theChild)
{
   QTag **array = new QTag *[mChildCount + 1];

   for(int i = 0; i < mChildCount; i ++)
      array[i] = mChildren[i];

   array[mChildCount++] = theChild;
   delete [] mChildren;
   mChildren = array;
}

bool QTag::Internal::removeChild(int index)
{
   QTag *child = takeChild(index);

   if(child)
   {
      delete child;
      return(true);
   }

   return(false);
}

QTag *QTag::Internal::takeChild(int index)
{
   if(index < 0 || index >= mChildCount)
      return(0);

   QTag *child  = mChildren[index];
   QTag **array = 0;

   if(mChildCount > 1)
   {
      array = new QTag *[mChildCount - 1];

      for(int i = 0, j = 0; i < mChildCount; i ++)
         if(i != index)
            array[j++] = mChildren[i];
   }

   --mChildCount;
   delete [] mChildren;
   mChildren = array;

   return(child);
}

/*****************************************************************************
 * Definition of Class: QTag
 *****************************************************************************/

QTag::QTag(const QString &tagName) :
   mInternal(new Internal(tagName))
{
}

QTag::~QTag()
{
   delete mInternal;
}

QTag *QTag::read(const QString &fileName, QString &errMsg)
{
   QTag *theTag = 0;

   QFile theFile(fileName);

   if(!theFile.open(IO_ReadOnly))
   {
      errMsg.sprintf("Unable to open file: [%s]", fileName.latin1());
   }
   else
   {
      QTextStream theStream(&theFile);
      QString contents = theStream.read();
      theFile.close();

      theTag = QTag::readFromText(contents, errMsg);
   }

   return(theTag);
}

// --------------------------------------
// --------------------------------------

class QScaleParser : public QXmlDefaultHandler
{
   public:
      QScaleParser();

      bool startElement(
         const QString &,
         const QString &,
         const QString &name,
         const QXmlAttributes &attrs
      );

      bool endElement(
         const QString &, const QString &, const QString &name
      );

      bool fatalError(const QXmlParseException &x);
      QString errorString();

      QTag *getTag();

   private:
      QTagStack mStack;
      QString   mErrMsg;
};

QScaleParser::QScaleParser()
{
   mStack.setAutoDelete(true);
}

bool QScaleParser::startElement(
   const QString &,
   const QString &,
   const QString &name,
   const QXmlAttributes &attrs
)
{
   QTag *theTag = new QTag(name);

   int n = attrs.length();

   for(int i = 0; i < n; i ++)
   {
      QString key   = attrs.localName(i);
      QString value = attrs.value(i);

      theTag->addAttribute(key, value);
   }

   mStack.push(theTag);

   return(true);
}

bool QScaleParser::endElement(
   const QString &, const QString &, const QString &
)
{
   // If we're not the top element in the stack, we need to add ourself
   // as a child to the QTag immediately above us

   if(mStack.count() > 1)
   {
      QTag *theTag = mStack.pop();
      QTag *top    = mStack.top();

      top->addChild(theTag);
   }

   return(true);
}

bool QScaleParser::fatalError(const QXmlParseException &x)
{
   mErrMsg.sprintf(
      "[line: %d, column: %d] - %s",
      x.lineNumber(), x.columnNumber(), x.message().latin1()
   );

   return(false);
}

QString QScaleParser::errorString()
{
   return(mErrMsg);
}

QTag *QScaleParser::getTag()
{
   return(mStack.pop());
}

QTag *QTag::readFromText(const QString &text, QString &errMsg)
{
   QScaleParser parser;

   QXmlInputSource src;
   src.setData(text);

   QXmlSimpleReader reader;
   reader.setContentHandler(&parser);
   reader.setErrorHandler(&parser);

   QTag *theTag = NULL;

   if(!reader.parse(src))
   {
      errMsg = parser.errorString();
   }
   else
   {
      theTag = parser.getTag();
   }

   return(theTag);
}

bool QTag::write(const QString &fileName, QString &errMsg) const
{
   const char *name = fileName.latin1();
   FILE *f          = 0;

   if((f = fopen(name, "wb")) == NULL)
   {
      errMsg.sprintf("Could not open file: [%s]", name);
      return(false);
   }

   mInternal->dump(f);

   fclose(f);

   return(true);
}

void QTag::writeToStdout() const
{
   mInternal->dump(stdout);
}

void QTag::accept(QTagVisitor &visitor) const
{
   visitor.begin(this);

   int count = getChildCount();

   for(int i = 0; i < count; i ++)
      getChild(i)->accept(visitor);

   visitor.end(this);
}

const QString &QTag::getAttribute(const QString &key) const
{
   return(mInternal->getAttribute(key));
}

bool QTag::addAttribute(const QString &key, const QString &value)
{
   return(mInternal->addAttribute(key, value));
}

bool QTag::removeAttribute(const QString &key)
{
   return(mInternal->removeAttribute(key));
}

void QTag::getKeys(AttributeKeyList &theList) const
{
   mInternal->getKeys(theList);
}

int QTag::getChildCount() const
{
   return(mInternal->getChildCount());
}

QTag *QTag::getChild(int index) const
{
   return(mInternal->getChild(index));
}

void QTag::addChild(QTag *theTag)
{
   mInternal->addChild(theTag);
}

bool QTag::removeChild(int index)
{
   return(mInternal->removeChild(index));
}

QTag *QTag::takeChild(int index)
{
   return(mInternal->takeChild(index));
}

const QString &QTag::name() const
{
   return(mInternal->mTagName);
}

int QTag::main(int argc, char **argv)
{
   if(argc < 2)
   {
      printf("Usage: %s {xmlFileName}\n", argv[0]);
      return(1);
   }

   QString fileName(argv[1]);
   QString errMsg;

   QTag *theTag = QTag::read(fileName, errMsg);

   if(!theTag)
   {
      printf("%s\n", errMsg.latin1());
   }
   else
   {
      theTag->writeToStdout();
      delete theTag;
   }

   return(0);
}

#ifdef __MAIN__

int main(int argc, char **argv)
{
   return(QTag::main(argc, argv));

}  /* main */

#endif
